/**
 * @file
 * Behaviours for profiler observer.
 */

// Namespace.
var qToolsProfilerPreviewObserver = {
  'loaded': false,
  'xhrWrapped': false,
  'requests': [],
  'watchers': [],
  'configuration': {},

};

(function (drupalSettings) {
  var observer = qToolsProfilerPreviewObserver;

  function log(message, value, info) {
    console.log("observer :: " + message, value, info)
  }

  observer.log = log;

  /**
   * Triggers set report content.
   */
  observer.setReportContent = function (report, content) {
    // Dispatch initial sync event.
    observer.dispatch('onSetReportContent', {
      'report': report,
      'content': content
    });
  };

  /**
   * Triggers select report event.
   */
  observer.reportSelected = function (report) {
    // Dispatch initial sync event.
    observer.dispatch('onSelectReport', {
      'report': report
    });
  };

  /**
   * Subscribe to observer events.
   */
  observer.subscribe = function (callbacks) {
    observer.watchers.push(callbacks);

    // Dispatch initial sync event.
    observer.dispatchOne(callbacks, 'onUpdateRequests', {
      'requests': observer.requests
    });
  };

  /**
   * Subscribe to observer events.
   */
  observer.dispatch = function (event, data) {
    // If we are not in the main windows we don't touch anything.
    if (self !== parent) {
      log('DISPATCH SKIP (not parent window)', event, data, data.length);
      return;
    }
    log('DISPATCH', event, data, data.length);
    for (var i = 0; i < observer.watchers.length; i++) {
      observer.dispatchOne(observer.watchers[i], event, data);
    }
  };

  /**
   * Subscribe to observer events.
   */
  observer.dispatchOne = function (callbacks, event, data) {
    if (callbacks[event]) {
      callbacks[event](data);
    }
  };

  /**
   * Hook every ajax request.
   */
  observer.wrapXhr = function () {
    if (this.xhrWrapped) {
      return;
    }

    this.xhrWrapped = true;
    if ('fetch' in window && self === parent) {
      var originalFetch = window.fetch;
      window.fetch = async (...args) => {
        var [resource, config] = args;
        if (!observer.isRequestTracked(resource)) {
          return await originalFetch(resource, config);
        }
        else {
          var requestName = config.method + ':' + resource;
          var trackingId = observer.pushRequest(requestName, null, 'xhr');
          var response = await originalFetch(resource, config);
          var requestId = response.headers.get('X-QTools-Profiler-RequestId');
          var requestSummary = response.headers.get('X-QTools-Profiler-RequestSummary');
          if (requestId) {
            observer.updateRequest(trackingId, requestId, requestSummary);
          }
          return response;
        }
      };
    }
    if (XMLHttpRequest && self === parent) {
      var origOpen = XMLHttpRequest.prototype.open;
      XMLHttpRequest.prototype.open = function() {

        // Skip technical requests.
        if (observer.isRequestTracked(arguments[1])) {
          var requestName = arguments[0] + ':' + arguments[1].split('?')[0];
          var trackingId = observer.pushRequest(requestName, null, 'xhr');
          this.addEventListener('load', function() {
            var requestId = this.getResponseHeader('X-QTools-Profiler-RequestId');
            var requestSummary = this.getResponseHeader('X-QTools-Profiler-RequestSummary');
            if (requestId) {
              observer.updateRequest(trackingId, requestId, requestSummary);
            }
          });
        }

        // Run original request.
        return origOpen.apply(this, arguments);
      };
    }
  };

  /**
   * Check if this request should be tracked.
   */
  observer.isRequestTracked = function(resource) {
    var request_is_tracked = true;

    // Do not track profiler own requests.
    if (resource.indexOf('/qtools_profiler/') !== -1) {
      request_is_tracked = false;
    }

    // Do not track external requests unless they match domain of current site.
    if (resource.indexOf('//') !== -1 && resource.indexOf('//' + window.location.hostname) === -1) {
      request_is_tracked = false;
    }

    return request_is_tracked;
  };

  /**
   * Push new request.
   */
  observer.pushRequest = function(path, requestId, type, requestSummary, timestamp, prepend) {
    var autotypes = [
      ['200 GET:', 'page'],
      ['200 POST:', 'xhr'],
      ['403', 'error'],
      ['404', 'error'],
      ['301', 'redirect'],
      ['302', 'redirect'],
      ['303 POST:', 'form'],
    ];

    if (!type) {
      type = autotypes.reduce((currentType, autotype) => {
        return path.indexOf(autotype[0]) !== -1 ? autotype[1] : currentType;
      }, 'unknown');
    }

    var trackingId = observer.requests.length;
    var summary = requestSummary ? JSON.parse(requestSummary) : [];
    var name = '../' + path.split('?')[0].replace(/\/$/,'').split('/').slice(-1)[0];
    var request = {
      'name': name,
      'path': path,
      'requestId': requestId,
      'summary': summary,
      'timestamp': timestamp,
      'trackingId': trackingId,
      'type': type
    };
    if (prepend) {
      observer.requests.unshift(request);
    }
    else {
      observer.requests.push(request);
    }

    observer.dispatch('onPushRequest', request);

    var reversed = observer.requests.slice();
    // reversed.reverse();
    observer.dispatch('onUpdateRequests', {
      'requests': reversed
    });
    return trackingId;
  };

  /**
   * Update request info.
   */
  observer.updateRequest = function(trackingId, requestId, requestSummary) {
    var summary = requestSummary ? JSON.parse(requestSummary) : [];
    observer.requests[trackingId].requestId = requestId;
    observer.requests[trackingId].summary = summary;

    observer.dispatch('onUpdateRequest', {
      'trackingId': trackingId,
      'requestId': requestId,
      'summary': summary,
    });
    observer.dispatch('onUpdateRequests', {
      'requests': observer.requests
    });
  };

  observer.loadStream = function(channelId, lastRequest, amount) {
    amount = amount || 15;
    lastRequest = lastRequest || -1;
    var url = '/admin/config/development/qtools_profiler/channels/' + [channelId, lastRequest, amount].join('/');
    console.log(url, channelId, lastRequest, amount);

    // Request profiler info for stream.
    var ajax = Drupal.ajax({
      url: url
    });
    ajax.execute();
  };

  observer.configuration = drupalSettings.qtools_profiler.configuration;
  observer.request_id = drupalSettings.qtools_profiler.request_id || null;

  // Some immediate actions.
  if (!observer.configuration.stream) {
    // wrap XHR.
    observer.wrapXhr();

    // Push main request for non stream mode.
    if (observer.request_id) {
      observer.pushRequest(window.location.href, observer.request_id, 'page');
    }
  }

})(drupalSettings);
