/**
 * @file
 * Behaviours for profiler wrapper.
 */

// Namespace.
var qToolsProfilerPreview = {
  'loaded': null,
  'defaultReport': null,
  'inspector': null,
  'requests': [],
  'reports': [],
  'iframeReceiver': null
};

(function () {
  var preview = qToolsProfilerPreview;
  var observer = qToolsProfilerPreviewObserver;

  function log(message, value, info) {
    var enabled = !!qtools.cookieGet('qToolsProfilerPreview.log');
    console.log("preview-silent :: " + message, value, info)
  }

  preview.log = log;

  /**
   * Sends updated requests list to inspector.
   */
  preview.onUpdateRequests = function (requests) {
    // Save request info locally for late inspector connections.
    preview.requests = requests;
    preview.sendRequestsList(requests);
  };

  /**
   * Sends report content to inspector.
   */
  preview.onSetReportContent = function (data) {
    preview.sendReportContent(data.report, data.content || {});
  };

  /**
   * Sends report content to inspector.
   */
  preview.sendForceReport = function (report) {
    preview.defaultReport = report;
    preview.sendMessage('FORCE_REPORT', {
      'report': report || ''
    });
  };


  /**
   * Sends report content to inspector.
   */
  preview.sendReportContent = function (report, content) {
    preview.sendMessage('REPORT_CONTENT_SET', {
      'report': report,
      'content': content
    });
  };

  /**
   * Sends request list to inspector.
   */
  preview.sendRequestsList = function (requests) {
    preview.sendMessage('REQUESTS_SET', requests);
  };

  /**
   * Sends configuration to inspector.
   */
  preview.sendConfiguration = function (configuration) {
    preview.sendMessage('CONFIGURATION', configuration);
  };

  /**
   * Sends reports list to inspector.
   */
  preview.sendReportsList = function (reports) {
    preview.sendMessage('REPORTS_SET', reports);
  };

  /**
   * Sends actions list to inspector.
   */
  preview.sendActionsList = function (actions) {
    preview.sendMessage('ACTIONS_SET', actions);
  };

  /**
   * Message handlers.
   */
  preview.handleMessage = {};

  /**
   * Handle connect message.
   */
  preview.handleMessage['CONNECT'] = function(params, e) {
    preview.inspector = params;
    log('Inspector connected :', preview.inspector);
    preview.sendConfiguration(preview.configuration);
    preview.sendRequestsList(preview.requests);
    preview.sendReportsList(preview.reports);
    preview.sendActionsList(preview.actions);
    preview.sendForceReport(preview.defaultReport);
  };

  /**
   * Handle report selection message.
   */
  preview.handleMessage['SELECT_REPORT'] = function(params, e) {
    preview.reportSelected(params);
  };

  /**
   * Handle action run message.
   */
  preview.handleMessage['RUN_ACTION'] = function(params, e) {
    preview.runAction(params);
  };

  /**
   * Handle action run message.
   */
  preview.handleMessage["UPDATE_CHANNEL_STREAM"] = function(params, e) {
    console.log('UPDATE_CHANNEL_STREAM');
    if (preview.configuration.channel_id) {
      const lastRequestId = preview.requests.reduce(function (max, request) {
        return Math.max(max, +request.requestId)
      }, 0);
      console.log(preview.configuration.channel_id, lastRequestId);
      observer.loadStream(preview.configuration.channel_id, lastRequestId)
    }
  };

  /**
   * Sends message to preview content frame.
   */
  preview.sendMessage = function (command, params) {
    var message = {
      'command': command,
      'params': params
    };
    if (preview.inspector) {
      message.target = preview.inspector.id;

      // We post this message directly to iframe container.
      if (preview.inspector.mode === 'iframe' && preview.iframeReceiver) {
        log('SEND', command, params);
        preview.iframeReceiver.postMessage(message, '*');
      }
      else {
        log('SEND failed', command, params);
      }
    }
    else {
      log('SEND skipped', command, params);
    }
  };

  /**
   * Select report.
   */
  preview.reportSelected = function(params) {
    observer.reportSelected(params.report);
  };

  /**
   * Run action.
   */
  preview.runAction = function(params) {
    let action = preview.actions.find((action) => action.actionId === params.actionId);
    if (action.type === 'request') {
      preview.runActionRequest(action, params);
    }
    console.log(action, params)
  };

  /**
   * Run button action.
   */
  preview.runActionRequest = function(action, params) {
    if (action.config.href) {
      var config = action.config;
      var url = config.href.replace('{value}', params.value);
      if (config.method === 'location') {
        location.href = url;
      }
      else if (config.method === 'ajax') {
        // Issue ajax call.
        var ajaxCall = Drupal.ajax({
          url: url
        });
        ajaxCall.execute();
      }
    }
  };

  /**
   * Handle pageLoad.
   */
  preview.handlePageLoad = function (context, settings) {
    if (!preview.loaded && settings.qtools_profiler) {
      preview.loaded = true;
      preview.configuration = settings.qtools_profiler.configuration || {};
      preview.reports = settings.qtools_profiler.reports || [];
      preview.actions = settings.qtools_profiler.actions || [];

      // Starts listening window messages.
      window.addEventListener('message', function(e) {
        if (e.data && e.data.target === 'preview') {
          if (preview.handleMessage[e.data.command]) {
            log('RECV ACCEPTED', e.data);
            preview.handleMessage[e.data.command](e.data.params, e);
          }
          else {
            log('RECV UNHANDLED', e.data);
          }
        }
        else {
          // log('RECV IGNORED', e);
        }
      });

      // Subscribe to observer.
      observer.subscribe({
        'onUpdateRequests': function (data) {
          preview.onUpdateRequests(data.requests);
        },
        'onSetReportContent': function (data) {
          preview.onSetReportContent(data);
        }
      });
    }
  };

  /**
   * Register behavior.
   */
  Drupal.behaviors.qToolsProfilerPreview = {
    attach: function (context, settings) {
      preview.handlePageLoad(context, settings);
    }
  };

})();