<?php

namespace Drupal\qtools_profiler\Controller;

use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormBuilderInterface;
use Drupal\Core\Render\HtmlResponse;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Routing\RouteMatch;
use Drupal\Core\Url;
use Drupal\qtools_profiler\Ajax\PreviewSummaryCommand;
use Drupal\qtools_profiler\PerformanceService;
use Drupal\qtools_profiler\ReportService;
use Drupal\qtools_profiler\Routing\ProfilerRoutes;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Class ChannelsReportController.
 *
 * @package Drupal\qtools_profiler\Controller
 */
class ChannelsReportController extends ControllerBase {

  /**
   * Drupal\qtools_profiler\PerformanceService.
   *
   * @var \Drupal\qtools_profiler\PerformanceService
   */
  protected $performanceService;

  /**
   * Drupal\qtools_profiler\ReportService.
   *
   * @var \Drupal\qtools_profiler\ReportService
   */
  protected $reportService;

  /**
   * Drupal\Core\Form\FormBuilderInterface.
   *
   * @var \Drupal\Core\Form\FormBuilderInterface
   */
  protected $formBuilder;

  /**
   * Drupal\Core\Extension\ModuleHandlerInterface.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    PerformanceService $performanceService,
    ReportService $reportService,
    FormBuilderInterface $formBuilder,
    ModuleHandlerInterface $moduleHandler,
    RendererInterface $renderer
  ) {
    $this->performanceService = $performanceService;
    $this->reportService = $reportService;
    $this->formBuilder = $formBuilder;
    $this->moduleHandler = $moduleHandler;
    $this->renderer = $renderer;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('qtools_profiler.performance'),
      $container->get('qtools_profiler.report'),
      $container->get('form_builder'),
      $container->get('module_handler'),
      $container->get('renderer')
    );
  }

  /**
   * Returns page content.
   *
   * @return array
   *   Page array.
   */
  public function pageBuild(Request $request) {

    // Get filter form.
    $build['form'] = $this->formBuilder->getForm('\Drupal\qtools_profiler\Form\ChannelsFilterForm');

    // Get report table.
    $filters = $this->reportService->getChannelsFilters();

    $header = $this->reportService->reportChannelsTableHeader($filters);
    $query = $this->reportService->reportChannelsBaseQuery($filters);
    $results = $query->execute()->fetchAll();
    $rows = $this->reportService->reportChannelsProcessResults($results, $filters);

    $build['table'] = [
      '#type' => 'table',
      '#rows' => $rows,
      '#header' => $header,
      '#cache' => [
        'max-age' => -1,
      ],
    ];

    $build['pager'] = [
      '#type' => 'pager',
    ];

    return $build;
  }

  /**
   * Returns channel stream page content.
   *
   * @return array
   *   Page array.
   */
  public function pageChannelViewBuild(Request $request) {

    // Get filter values.
    $route = RouteMatch::createFromRequest($request);
    $channel_id = $route->getParameter('id');

    // List preview reports, exclude static one.
    $preview_reports = $this->performanceService->listProfilerReports($attachments, TRUE);

    // Get filter form.
    $build['content'] = [
      '#type' => 'theme',
      '#theme' => 'qtools_profiler_channel_stream',
      '#attached' => [
        'drupalSettings' => [
          'qtools_profiler' => [
            'configuration' => [
              'channel_id' => $channel_id,
              'stream' => TRUE,
              'fullscreen' => TRUE,
              'preview_app_url' => Url::fromRoute(ProfilerRoutes::PREVIEW_APP_IFRAME_ROUTE)->toString(),
            ],
            'reports' => $preview_reports,
          ],
        ],
      ],
    ];

    return $build;
  }

  /**
   * Returns channel stream requests.
   */
  public function ajaxChannelStream(Request $request) {

    // Get filter values.
    $route = RouteMatch::createFromRequest($request);
    $channel_id = $route->getParameter('id');
    $last_request_id = $route->getParameter('last_request_id');
    $count = $route->getParameter('count');

    // List preview reports, exclude static one.
    $request_ids = $this->performanceService->listChannelRequests($channel_id, $last_request_id, $count);

    // Push info about all requests.
    $response = new AjaxResponse();

    foreach ($request_ids as $rid => $timestamp) {
      $name = $this->performanceService->getRequestName($rid);
      $summary = $this->performanceService->getRequestSummary($rid);
      $response->addCommand(new PreviewSummaryCommand($rid, $rid, $timestamp, $name, $summary));
    }
    return $response;
  }

}
