<?php

namespace Drupal\qtools_profiler\EventSubscriber;

use Drupal\Core\Render\HtmlResponse;
use Drupal\Core\Site\Settings;
use Drupal\Core\Url;
use Drupal\dynamic_page_cache\EventSubscriber\DynamicPageCacheSubscriber;
use Drupal\qtools_common\Utility\Slug;
use Drupal\qtools_profiler\PerformanceService;
use Drupal\qtools_profiler\Routing\ProfilerRoutes;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\Event\TerminateEvent;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\HttpKernel\Event\RequestEvent;

/**
 * Force anonymous user redirect to login page.
 */
class OnKernelEvents implements EventSubscriberInterface {

  /**
   * Drupal\qtools_profiler\PerformanceService.
   *
   * @var \Drupal\qtools_profiler\PerformanceService
   */
  protected $performanceService;

  const PREVIEW_ACTION_WIDGET_TEXTFIELD = 'textfield';
  const PREVIEW_ACTION_WIDGET_BUTTON = 'button';

  const PREVIEW_ACTION_TYPE_REQUEST = 'request';

  /**
   * {@inheritdoc}
   */
  public function __construct(
    PerformanceService $performanceService,
  ) {
    $this->performanceService = $performanceService;
  }

  /**
   * Handles kernel terminate event.
   */
  public function onKernelTerminate(TerminateEvent $event) {
    if ($this->performanceService->monitoringActive()) {
      $this->performanceService->monitoringFinish($event);
    }
  }

  /**
   * Handles kernel response event.
   */
  public function onKernelResponse(ResponseEvent $event) {
    if (!$this->performanceService->monitoringActive()) {
      return;
    }

    $request_id = $this->performanceService->getRequestId();
    $this->performanceService->monitoringComplete($event);

    // Inject tracking id into response regardless of caching policy.
    $response = $event->getResponse();
    if ($response instanceof HtmlResponse) {
      // Get current attachment array.
      $attachments = $this->attachProfilerPreview($response, $request_id);
      if (empty($attachments)) {
        return;
      }

      // Get DC result.
      $dynamic_cache_status = $response->headers->get(DynamicPageCacheSubscriber::HEADER, 'UND');
      $page_cache_status = $response->headers->get('X-Drupal-Cache', 'UND');
      $attachments['html_head'][] = [
        [
          '#type' => 'html_tag',
          '#tag' => 'meta',
          '#attributes' => [
            'name' => 'qtools-profiler-cache-status',
            'id' => 'qtools-profiler-cache-status',
            'content' => $page_cache_status . ":" . $dynamic_cache_status,
          ],
          '#weight' => 1,
        ],
        'qtools_profiler_page_cache',
      ];

      $response->setAttachments($attachments);
    }
  }

  /**
   * Generates a unique URL for a user to log user programmatically.
   *
   * @see user_pass_reset_url()
   *
   * @return string
   *   A unique URL that provides a one-time log in for the user.
   */
  protected function userLoginUrl($account, $options = []): string {
    $timestamp = \Drupal::time()->getRequestTime();
    $langcode = $options['langcode'] ?? $account->getPreferredLangcode();
    return Url::fromRoute('qtools_profiler.user_login',
      [
        'uid' => $account->id(),
        'timestamp' => $timestamp,
        'hash' => user_pass_rehash($account, $timestamp),
      ],
      [
        'absolute' => TRUE,
        'language' => \Drupal::languageManager()->getLanguage($langcode),
      ]
    )->toString();
  }

  /**
   * Collects preview actions.
   */
  protected function getPreviewActions(): array {
    $preview_actions = [];

    // Add links fom settings.inc.
    $development_links = Settings::get('qtools_profiler_development_links', []);
    foreach ($development_links as $development_link) {
      $preview_actions[] = [
        'actionId' => Slug::slugify($development_link[0]),
        'type' => static::PREVIEW_ACTION_TYPE_REQUEST,
        'widget' => static::PREVIEW_ACTION_WIDGET_BUTTON,
        'label' => $development_link[0],
        'config' => [
          'method' => 'location',
          'href' => $development_link[1],
          'target' => $development_link[2] ?? '_self',
        ],
      ];
    }

    // Add actions to quickly login as dev users on local.
    $development_users = Settings::get('qtools_profiler_development_users', []);
    foreach ($development_users as $development_user) {
      /** @var \Drupal\user\Entity\User $account */
      $account = user_load_by_name($development_user);

      // User does not exists.
      if (empty($account)) {
        continue;
      }

      // User already logged in.
      if ($account->id() === \Drupal::currentUser()->id()) {
        continue;
      }

      $preview_actions[] = [
        'actionId' => 'auto_login_' . $account->id(),
        'type' => static::PREVIEW_ACTION_TYPE_REQUEST,
        'widget' => static::PREVIEW_ACTION_WIDGET_BUTTON,
        'label' => $account->getAccountName(),
        'config' => [
          'method' => 'location',
          'href' => $this->userLoginUrl($account),
        ],
      ];
    }

    // List actions.
    if (\Drupal::currentUser()->isAuthenticated()) {
      $preview_actions[] = [
        'actionId' => 'logout',
        'type' => static::PREVIEW_ACTION_TYPE_REQUEST,
        'widget' => static::PREVIEW_ACTION_WIDGET_BUTTON,
        'label' => t('Logout'),
        'config' => [
          'method' => 'location',
          'href' => Url::fromRoute('user.logout')->setAbsolute()->toString(),
        ],
      ];
    }

    return $preview_actions;
  }

  /**
   * Inject profiler resources into page attachments.
   *
   * @return array
   *   Modified attachments array.
   */
  protected function attachProfilerPreview(HtmlResponse $response, $request_id): ?array {
    // Iframe preview only available in development mode.
    $iframe_preview_mode = Settings::get('qtools_profiler_development_mode') &&
      $this->performanceService->confGet('preview_mode') == PerformanceService::PREVIEW_MODE_IFRAME;

    // Only one preview mode is currently available.
    if (!$iframe_preview_mode) {
      return NULL;
    }

    /** @var \Drupal\qtools_profiler\Access\ProfilerPreviewAccessCheck $qToolsProfilerPreviewAccess */
    $qToolsProfilerPreviewAccess = \Drupal::service('access_check.qtools_profiler_preview');
    $access_iframe_review = $qToolsProfilerPreviewAccess->access(\Drupal::currentUser());
    if (!$access_iframe_review->isAllowed()) {
      return NULL;
    }
    $attachments = $response->getAttachments();

    $attachments['library'][] = 'qtools_profiler/profiler-preview-iframe';
    $attachments['drupalSettings']['qtools_profiler']['request_id'] = $request_id;

    // Get preview actions.
    $preview_actions = $this->getPreviewActions();

    // Default scales for cache contexts.
    $scales = [
      [
        'languages:language_interface$',
        count(\Drupal::languageManager()->getLanguages()),
      ],
      [
        'languages:language_content$',
        count(\Drupal::languageManager()->getLanguages()),
      ],
      ['session.exists', 2],
      ['cookies:', 2],
      ['qtools_cache_profiler_active', 1],
      ['theme$', 1],
      ['timezone$', 1],
      ['user.roles:', 2],
      ['user.node_grants:', 2],
      ['user.permissions$', 10],
      ['user$', 1000],
      ['url$', 1000],
      ['url.site$', 1],
      ['url.path.is_front$', 2],
      ['route$', 100],
      ['.*', 10],
    ];

    // Allow other modules to alter things.
    $moduleHandler = \Drupal::moduleHandler();

    $moduleHandler->alter('qtools_profiler_contexts_scales', $scales);
    $attachments['drupalSettings']['qtools_profiler']['contexts_scales'] = $scales;

    $moduleHandler->alter('qtools_profiler_preview_actions', $preview_actions);
    $attachments['drupalSettings']['qtools_profiler']['actions'] = $preview_actions;

    // List preview reports.
    $preview_reports = $this->performanceService->listProfilerReports($attachments);

    $attachments['drupalSettings']['qtools_profiler']['reports'] = $preview_reports;
    $attachments['drupalSettings']['qtools_profiler']['configuration'] = [
      'stream' => FALSE,
      'preview_app_url' => Url::fromRoute(ProfilerRoutes::PREVIEW_APP_IFRAME_ROUTE)->toString(),
    ];

    return $attachments;
  }

  /**
   * Handles kernel request event.
   */
  public function onKernelRequestEarly(RequestEvent $event) {
    // Ensure cookie client.
    $this->performanceService->monitoringEnsureCookie($event);

    // Start monitoring if allowed.
    if ($this->performanceService->monitoringAllowed($event)) {
      $this->performanceService->monitoringStart($event);
    }
  }

  /**
   * Handles kernel request event.
   */
  public function onKernelRequestWithRoute(RequestEvent $event) {
    if ($this->performanceService->monitoringActive()) {
      $route = $event->getRequest()->attributes->get('_route');
      if (!empty($route)) {
        $this->performanceService->setRouteAndUser($route);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events = [];
    $events[KernelEvents::TERMINATE][] = 'onKernelTerminate';

    // Run after DynamicPageCacheSubscriber, priority: 27.
    $events[KernelEvents::RESPONSE][] = ['onKernelResponse', 26];
    $events[KernelEvents::REQUEST][] = ['onKernelRequestEarly', 10000];
    $events[KernelEvents::REQUEST][] = ['onKernelRequestWithRoute', 31];
    return $events;
  }

}
