<?php

namespace Drupal\qtools_profiler\Form;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\qtools_profiler\PerformanceService;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configuration form for QTools profiler.
 */
class StatusConfigurationForm extends FormBase {

  /**
   * Drupal\qtools_profiler\PerformanceService.
   *
   * @var \Drupal\qtools_profiler\PerformanceService
   */
  protected $performanceService;

  /**
   * Drupal\Core\Session\AccountProxyInterface.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $accountProxy;

  /**
   * Drupal\Core\Messenger\MessengerInterface.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Config Factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    PerformanceService $performanceService,
    AccountProxyInterface $accountProxy,
    MessengerInterface $messenger,
    ConfigFactoryInterface $configFactory,
  ) {
    $this->performanceService = $performanceService;
    $this->accountProxy = $accountProxy;
    $this->messenger = $messenger;
    $this->configFactory = $configFactory;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('qtools_profiler.performance'),
      $container->get('current_user'),
      $container->get('messenger'),
      $container->get('config.factory'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'status_configuration_form';
  }

  /**
   * Ajax callback for key generation.
   */
  public function ajaxGenerateAccessKey(array &$form, FormStateInterface $form_state) {
    $errors = $form_state->getErrors();
    if (empty($errors['access_key_ttl'])) {
      $ttl = $form_state->getValue('access_key_ttl');
      $label = $form_state->getValue('access_key_label');
      $access_cid = $this->performanceService->generateNewClientId($label, $ttl);
      $form['conf']['access_cookie']['access_cookie_cid']['#value'] = 'document.cookie = "qtools_profiler_client=' . $access_cid . ';path=/;Secure;max-age=' . $ttl . '";';
    }
    return $form['conf']['access_cookie']['access_cookie_cid'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $conf = $this->performanceService->confGet();

    $form['conf'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Configuration'),
      '#tree' => TRUE,
    ];

    $form['conf']['enabled'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enabled'),
      '#default_value' => !empty($conf['enabled']),
    ];
    $form['conf']['preview_theme'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Preview theme'),
      '#default_value' => $conf['preview_theme'] ?: $this->configFactory->get('system.theme')->get('admin'),
    ];
    $form['conf']['preview_mode'] = [
      '#type' => 'select',
      '#title' => $this->t('Preview mode'),
      '#description' => $this->t('This setting only available in "Development" mode.'),
      '#default_value' => $conf['preview_mode'],
      '#options' => [
        PerformanceService::PREVIEW_MODE_NONE => $this->t('None'),
        PerformanceService::PREVIEW_MODE_IFRAME => $this->t('Iframe'),
      ],
      '#disabled' => empty($conf['developer_mode']),
    ];
    $form['conf']['developer_mode'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Development mode'),
      '#description' => $this->t('This setting can only be changed in code $settings["qtools_profiler_development_mode"] = TRUE/FALSE;.'),
      '#default_value' => $conf['developer_mode'],
      '#disabled' => TRUE,
    ];

    // Access settings.
    $form['conf']['client_id_auto'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Automatically generate client Ids'),
      '#description' => $this->t('This setting only available in "Development" mode.'),
      '#default_value' => !empty($conf['client_id_auto']),
      '#disabled' => empty($conf['developer_mode']),
    ];

    $form['conf']['access_cookie'] = [
      '#tree' => FALSE,
      '#type' => 'details',
      '#title' => $this->t('Personal access by cookie key'),
      '#collapsed' => TRUE,
      '#collapsible' => TRUE,
      '#states' => [
        'visible' => [
          ':input[name="conf[client_id_auto]"]' => ['checked' => FALSE],
        ],
      ],
    ];
    $form['conf']['access_cookie']['access_key_label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Key label'),
      '#default_value' => $this->accountProxy->getAccountName(),
      '#description' => $this->t('Key label or user name.'),
    ];
    $form['conf']['access_cookie']['access_key_ttl'] = [
      '#type' => 'number',
      '#min' => PerformanceService::MIN_SESSION_TTL,
      '#max' => PerformanceService::MAX_SESSION_TTL,
      '#title' => $this->t('Key ttl, seconds'),
      '#default_value' => PerformanceService::DEFAULT_SESSION_TTL,
      '#description' => $this->t('You can not retract access keys validity so be mindful with max ttl.'),
    ];
    $form['conf']['access_cookie']['generate_access_key'] = [
      '#type' => 'button',
      '#value' => $this->t('Generate'),
      '#ajax' => [
        'callback' => '::ajaxGenerateAccessKey',
        'disable-refocus' => TRUE,
        'event' => 'click',
        'wrapper' => 'target-access_cookie_cid',
        'progress' => [
          'type' => 'throbber',
          'message' => $this->t('Generating access key...'),
        ],
      ],
    ];

    $form['conf']['access_cookie']['access_cookie_cid'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Cookie client id'),
      '#description' => $this->t('Copy content of this field into console to start session in client brwoser.'),
      '#default_value' => '',
      '#disabled' => TRUE,
      '#prefix' => '<div id="target-access_cookie_cid">',
      '#suffix' => '</div>',
    ];

    $form['conf']['cleanup'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Cleanup ttl'),
      '#description' => $this->t('For how long keep records in db.'),
      '#default_value' => $conf['cleanup'],
    ];
    $form['conf']['flush'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Flush logs at day start'),
      '#description' => $this->t('To prevent possible database creep its good to truncate tables from time to time, if enabled, flush operation will be done during first cron each day.'),
      '#default_value' => $conf['flush'],
    ];

    // Monitoring settings.
    $form['conf']['monitoring'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Monitoring'),
      '#tree' => TRUE,
    ];
    $form['conf']['monitoring']['rules'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Rules to match requests to monitor'),
      '#default_value' => $conf['monitoring']['rules'],
      '#description' => $this->t('format is {fraction}|[path/role/uid]={value}; example: 1|path=* (track each request for all pages).'),
    ];
    $form['conf']['monitoring']['exclude'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Rules to exclude requests from monitor'),
      '#default_value' => $conf['monitoring']['exclude'],
      '#description' => $this->t('format is 1|[path/role/uid]={value}; example: <br>1|path=*/admin/* <br>1|path=*/node/*/* <br>1|path=*/taxonomy/term/*/* <br>1|path=*/user/*/* <br>(exclude all requests from */admin/* pages).'),
    ];
    $form['conf']['monitoring']['dblogging'] = [
      '#type' => 'select',
      '#title' => $this->t('Monitor database queries'),
      '#options' => [
        PerformanceService::DB_MODE_NONE => $this->t('None'),
        PerformanceService::DB_MODE_TOTALS => $this->t('Totals'),
      ],
      '#description' => $this->t('Specify to which extent we should monitor database queries.'),
      '#default_value' => $conf['monitoring']['dblogging'] ?? PerformanceService::DB_MODE_NONE,
    ];

    // Profiler settings.
    $form['conf']['profiling'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Profiling'),
      '#tree' => TRUE,
    ];
    $profiler_extensions = $this->performanceService->getProfilerOptions();
    $form['conf']['profiling']['extension'] = [
      '#type' => 'select',
      '#title' => $this->t('Select profiler'),
      '#options' => $profiler_extensions,
      '#default_value' => $conf['profiling']['extension'],
    ];

    $form['conf']['profiling']['profilers'] = $this->performanceService->getProfilerOptionsForm();
    foreach ($form['conf']['profiling']['profilers'] as $key => $value) {
      $form['conf']['profiling']['profilers'][$key]['#states'] = [
        'visible' => [
          ':input[name="conf[profiling][extension]"]' => ['value' => $key],
        ],
      ];
    }

    $form['conf']['profiling']['rules'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Rules to match requests to profile'),
      '#default_value' => $conf['profiling']['rules'],
      '#description' => $this->t('format is {fraction};[path|role|uid]={value}'),
    ];
    $form['conf']['profiling']['cookie'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Profiler cookie value'),
      '#description' => $this->t('Cookie name is "qtools_profiler", enter "+" to generate random one. If less than 10 symbols will be ignored.'),
      '#default_value' => $conf['profiling']['cookie'],
    ];

    $form['flush'] = [
      '#type' => 'details',
      '#title' => $this->t('Flush all data'),
      '#collapsed' => TRUE,
      '#collapsible' => TRUE,
    ];

    $form['flush']['flush_all'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Flush all stats'),
      '#description' => $this->t('This action cannot be undone.'),
      '#default_value' => FALSE,
    ];

    $form['flush']['flush_all'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Flush all stats'),
      '#description' => $this->t('This action cannot be undone.'),
      '#default_value' => FALSE,
    ];

    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Submit'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $conf = $form_state->getValue('conf');

    // Check if we need to generate random profiler cookie.
    if ($conf['profiling']['cookie'] == '+') {
      $conf['profiling']['cookie'] = uniqid('pc', TRUE);
    }
    $this->performanceService->confSet($conf);
    $this->messenger->addMessage($this->t('Changes saved.'));

    // Check for log clear.
    $flush = $form_state->getValue('flush_all');
    if (!empty($flush)) {
      $this->performanceService->flush();
      $this->messenger->addMessage($this->t('All data deleted.'));
    }
  }

}
